import {
  Directive,
  ElementRef,
  Input,
  NgZone,
} from "@angular/core";
import {
  isFunction,
  isNullOrUndefined,
  isTruthy,
} from "../../tools";

@Directive({
  selector: "[solidifyAutoFocus]",
})
export class AutoFocusDirective {

  // Focus timeout

  private _focusTimeout: number | undefined;

  // Auto focus

  private _autoFocus: boolean = false;

  @Input("solidifyAutoFocus")
  set autoFocus(value: boolean) {
    this._autoFocus = isTruthy(value);
    this._onAutoFocusChange();
  }

  get autoFocus(): boolean {
    return this._autoFocus;
  }

  // Native element

  get nativeElement(): HTMLElement | null {
    return this._elementRef.nativeElement;
  }

  // Constructor

  constructor(private readonly _elementRef: ElementRef,
              private readonly _ngZone: NgZone) {
  }

  // Is auto focusing

  isAutoFocusing(): boolean {
    return !isNullOrUndefined(this._focusTimeout);
  }

  // Focus

  private _focus(): boolean {
    const nativeElement = this.nativeElement;
    if (nativeElement && isFunction(nativeElement.focus)) {
      nativeElement.focus();
    } else {
      return false;
    }
  }

  // Set focus timeout

  private _setFocusTimeout(): number {
    this.clearFocusTimeout();
    this._focusTimeout = this._ngZone.runOutsideAngular(() => setTimeout(() => {
      this._focus();
      this._focusTimeout = undefined;
    }));
    return this._focusTimeout;
  }

  // Clear focus timeout

  clearFocusTimeout(): boolean {
    if (this._focusTimeout) {
      this._ngZone.runOutsideAngular(() => clearTimeout(this._focusTimeout));
      this._focusTimeout = undefined;
      return true;
    }
    return false;
  }

  // Update from auto focus

  private _onAutoFocusChange(autoFocus: boolean = this._autoFocus): void {
    if (autoFocus) {
      if (!this.isAutoFocusing()) {
        this._setFocusTimeout();
      }
    } else {
      this.clearFocusTimeout();
    }
  }

}
